//
//from firebird - overfly
//

#include <stdio.h>
//#define SEEK_SET 0
#include "tap.h"
#include "HDDIO.h"


typedef struct
{
	bool			bTurbolader;
	TYPE_File		*file;
	dword			sector;
	tTAP_Hdd_ReadSectorDMA  ReadSectorDMA;
} tHdd;

tHdd hdd;
static TYPE_Fileex fpx;
static byte * sectorBuffer;

void Hdd_Init()
{
	byte buffer[512];
//TAP_Print("\nHdd_Init()");
	hdd.ReadSectorDMA = (tTAP_Hdd_ReadSectorDMA) TAP_GetSystemProc(oTAP_Hdd_ReadSectorDMA);
	hdd.ReadSectorDMA(0, 1, buffer);
//dump(buffer, 0x80);
	fpx.SectorsPerCluster = (buffer[0x22] << 8) | buffer[0x23];
//TAP_Print("\nSectorsPerCluster=0x%x %d", fpx.SectorsPerCluster, fpx.SectorsPerCluster);
}

//
//chase through clusters, to find actual sector of data
//will get slower as we get near end of big file ...
//data starts in 2nd cluster
//
dword Hdd_CalcAbsSectorFromFAT(dword relSector)
{
	dword absCluster;

	absCluster = fpx.startCluster;
//TAP_Print("\nHdd_CalcAbsSectorFromFAT(), fpx.startCluster=0x%x", fpx.startCluster);
	if (absCluster > 131071) {        //0x1FFFF
        TAP_Print("\nerror: Hdd_CalcAbsSectorFromFAT(), fpx.startCluster=0x%x %d", fpx.startCluster, fpx.startCluster);
	    return 0;
	}

	while (relSector >= fpx.SectorsPerCluster) {
		absCluster = fpx.FAT[absCluster];
		if (absCluster > 131071) {
            TAP_Print("\nerror:absCluster=%x", absCluster);
		    return 0;
		}
		relSector -= fpx.SectorsPerCluster;
	}
//final cluster
//    TAP_Print("\nabsCluster=%x, relSector=%x, abssector=%x", absCluster, relSector, fpx.SectorsPerCluster * (absCluster+1)+relSector);
	return (fpx.SectorsPerCluster * (absCluster+1)+relSector);
}

//

bool Hdd_Fopen(TYPE_File *file, bool turbo)
{
	int i;
	byte *buf;

//    TAP_Print("\nHdd_Fopen()");
	hdd.file = file;
	if (hdd.file == NULL) {
        TAP_Print("\nHdd_Fopen(), hdd.file=%x", hdd.file);
	    return FALSE;
	}
//	if (!turbo || (hdd.file->size>>10) < 4090000) {                          //4gb ?
//	if (!turbo || ((*(dword *)(4+(int)&hdd.file->size))>>10) < 4090000) {
	if (!turbo) {
		hdd.bTurbolader = FALSE;
//        TAP_Print("\nHdd_Fopen(), hdd.file->size=%x", *(dword *)(4+(int)&hdd.file->size));
//        TAP_Print("\nHdd_Fopen(), hdd.bTurbolader=%x", hdd.bTurbolader);
		return TRUE;
	}
	hdd.bTurbolader = TRUE;
    memcpy(&fpx, hdd.file, sizeof(TYPE_File));
	buf = ((byte *) fpx.FAT)+131072;
	hdd.ReadSectorDMA(256, 768, buf);           //98Kdwords
//dump(buf, 0x200);
	for (i=0; i<131072; i++) {
		fpx.FAT[i] = (*buf << 16) | (*(buf+1) << 8) | *(buf+2);
		buf+=3;
	}
	hdd.sector = 0;
//    TAP_Print("\nHdd_Fopen(), hdd.bTurbolader=%x", hdd.bTurbolader);
//    TAP_Print("\nHdd_Fopen(), hdd.sector=%x", hdd.sector);
	return TRUE;
}

//
//
//need to get cluster by cluster ...
//
dword Hdd_Fread(dword size, byte *buffer)
{
	dword absSector, relSector, intlSector, num_sectors, remBytes;
	int ret;

	if (!hdd.bTurbolader) {
		ret = TAP_Hdd_Fread(buffer, size, 1, hdd.file);
        return ret;
    }

	sectorBuffer = buffer;
	relSector = hdd.sector;
    intlSector = relSector % fpx.SectorsPerCluster;
    remBytes = size;

    //1st partial cluster
    num_sectors = fpx.SectorsPerCluster - intlSector;
    if (num_sectors >= (remBytes>>9)) num_sectors = 0;    //don;t bother if only 1 cluster
    if (num_sectors) {
	    absSector = Hdd_CalcAbsSectorFromFAT(relSector);
	    if (absSector == 0) {
	        return 0;
	    }
	    ret = hdd.ReadSectorDMA (absSector, num_sectors, sectorBuffer);
  	    relSector += num_sectors;
    	sectorBuffer += (num_sectors<<9);
  	    remBytes -= (num_sectors<<9);
    }
    //whole clusters
    num_sectors = fpx.SectorsPerCluster;
    while (remBytes > (fpx.SectorsPerCluster<<9)) {
	    //increasingly slow since it scans the whole fat
	    //better to find start point, then find next
	    absSector = Hdd_CalcAbsSectorFromFAT(relSector);
	    if (absSector == 0) {
	        return 0;
	    }
	    ret = hdd.ReadSectorDMA (absSector, num_sectors, sectorBuffer);
  	    relSector += num_sectors;
    	sectorBuffer += (num_sectors<<9);
  	    remBytes -= (num_sectors<<9);
    }
    //final partial cluster
    num_sectors = remBytes>>9;
    if (num_sectors == 0) num_sectors = 1;  //round up the final sector ... and hope the users buffer is big enough
    if (num_sectors) {
	    absSector = Hdd_CalcAbsSectorFromFAT(relSector);
	    if (absSector == 0) {
	        return 0;
	    }
	    ret = hdd.ReadSectorDMA (absSector, num_sectors, sectorBuffer);
  	    relSector += num_sectors;
    	sectorBuffer += (num_sectors<<9);
  	    remBytes -= (num_sectors<<9);
    }

    hdd.sector += size>>9;
	return size;
}

//

void Hdd_Fclose()
{
    TAP_Hdd_Fclose(hdd.file);
}

//
// Seek Genauigkeit = 1 Sector = 512 Byte
//

dword Hdd_Fseek(dword position)
{
	if (!hdd.bTurbolader)
		return TAP_Hdd_Fseek(hdd.file, position<<9, SEEK_SET);      //just sets hdd.file.currentPos=(position<<9)
	hdd.sector = position;
	return 0;
}
